#include "stdafx.h"
#include "RunnerDB.h"
#include "xmlparser.h"

#include "io.h"
#include "fcntl.h"
#include "sys/stat.h"
#include "meos_util.h"
#include "oDataContainer.h"

RunnerDB::RunnerDB(oEvent *oe_): oe(oe_)
{
}

RunnerDB::~RunnerDB(void)
{
}

RunnerDBEntry::RunnerDBEntry()
{
  memset(this, 0, sizeof(RunnerDBEntry));  
}

void RunnerDBEntry::getName(string &n) const
{
  n=name;
}

void RunnerDBEntry::setName(const char *n) 
{
  memcpy(name, n, min<int>(strlen(n)+1, baseNameLength));
  name[baseNameLength-1]=0;
}

string RunnerDBEntry::getGivenName() const
{
  string Name(name);
	return trim(Name.substr(0, Name.find_first_of(' ')));
}

string RunnerDBEntry::getFamilyName() const
{
  string Name(name);
  int k=Name.find_first_of(' ');

  if(k==string::npos)
    return "";
  else return trim(Name.substr(k, string::npos));
}

RunnerDBEntry *RunnerDB::addRunner(const char *name, 
                                   int club, int card)
{
  rdb.push_back(RunnerDBEntry());
  RunnerDBEntry &e=rdb.back();
  e.cardNo=card;
  e.clubNo=club;
  e.setName(name);

  if (!check(e) )
    rdb.pop_back();
  else
    rhash[card]=rdb.size()-1;

  return &e;
}

void RunnerDB::addClub(oClub &c)
{
  cdb.push_back(c);
  chash[c.getId()]=cdb.size()-1;
}

void RunnerDB::importClub(oClub &club)
{
  pClub pc = getClub(club.getId());

  if (pc && !pc->sameClub(club)) {
    //The new id is used by some other club.
    //Remap old club first

    int oldId = pc->getId(); 
    int newId = chash.rbegin()->first + 1;

    for (size_t k=0; k<rdb.size(); k++)
      if (rdb[k].clubNo == oldId)
        rdb[k].clubNo = newId;

    chash[newId] = chash[oldId];
    pc = 0;
  }

  if (pc == 0) {
    // Try to find the club under other id
    pc = getClub(club.getName());

    if ( pc!=0 ) {
      // If found under different id, remap to new id
      int oldId = pc->getId(); 
      int newId = club.getId();

      for (size_t k=0; k<rdb.size(); k++)
        if (rdb[k].clubNo == oldId)
          rdb[k].clubNo = newId;
      
      pClub base = &cdb[0];
      int index = (size_t(pc) - size_t(base)) / sizeof(oClub);
      chash[newId] = index;
    }
  }

  if ( pc )
    *pc = club; // Update old club
  else {
    // Completely new club
    chash [club.getId()] = cdb.size();
    cdb.push_back(club);    
  }
}

void RunnerDB::compactifyClubs()
{
  chash.clear();

  map<int, int> clubmap;  
  for (size_t k=0;k<cdb.size();k++) 
    clubmap[cdb[k].getId()] = cdb[k].getId();

  for (size_t k=0;k<cdb.size();k++) {
    oClub &ref = cdb[k];    
    vector<int> compacted;
    for (size_t j=k+1;j<cdb.size(); j++) {
      if(_stricmp(ref.getName().c_str(), 
              cdb[j].getName().c_str())==0) 
        compacted.push_back(j);
    }

    if (!compacted.empty()) {
      int ba=ref.getDI().getDataAmountMeasure();  
      oClub *best=&ref;
      for (size_t j=0;j<compacted.size();j++) {
        int nba=cdb[compacted[j]].getDI().getDataAmountMeasure();       
        if (nba>ba) {
          best = &cdb[compacted[j]];
          ba=nba;
        }
      }
      swap(ref, *best);

      //Update map
      for (size_t j=0;j<compacted.size();j++) {
        clubmap[cdb[compacted[j]].getId()] = ref.getId();
      }      
    }
  }
}


bool RunnerDB::getRunner(int card, string &name, string &club) const
{
  map<int,int>::const_iterator it = rhash.find(card);

  if (it!=rhash.end()) {
    int i=it->second;
    rdb[i].getName(name);
    getClub(rdb[i].clubNo, club);
    return true;
  }
  return false;
}

RunnerDBEntry *RunnerDB::getRunner(int card) const
{
  map<int,int>::const_iterator it = rhash.find(card);

  if (it!=rhash.end())
    return (RunnerDBEntry *)&rdb[it->second];

  return 0;
}

bool RunnerDB::getClub(int clubId, string &club) const
{
  map<int,int>::const_iterator it = chash.find(clubId);

  if (it!=chash.end()) {
    int i=it->second;
    club=cdb[i].getName();
    return true;
  }
  return false;
}

oClub *RunnerDB::getClub(int clubId) const
{
  map<int,int>::const_iterator it = chash.find(clubId);

  if (it!=chash.end())
    return pClub(&cdb[it->second]);
  
  return 0;
}

oClub *RunnerDB::getClub(const string &name) const
{
  string tName=trim(name);
  // Do a linear search
  for (size_t k=0; k<cdb.size(); k++)
    if (_stricmp(cdb[k].getName().c_str(), tName.c_str())==0)
      return pClub(&cdb[k]);

  return 0;
}

void RunnerDB::saveClubs(const char *file)
{
	xmlparser xml;

	if(!xml.OpenOutput(file, "meosclubs"))
    throw std::exception((string("Fel. Kan inte ppna: ") + file).c_str());

  vector<oClub>::iterator it;	

	xml.StartTag("ClubList");

	for (it=cdb.begin(); it != cdb.end(); ++it)
		it->Write(xml);

	xml.EndTag();

	xml.CloseOut();
}

void RunnerDB::saveRunners(const char *file)
{
  if (rdb.empty())
    return;

  int f=-1;
  _sopen_s(&f, file, _O_BINARY|_O_CREAT|_O_TRUNC|_O_WRONLY, 
            _SH_DENYWR, _S_IREAD|_S_IWRITE);
 
  if (f!=-1) {
    _write(f, &rdb[0], rdb.size()*sizeof(RunnerDBEntry));
    _close(f);
  }
  else throw std::exception("Could not save runner database.");
}

void RunnerDB::loadClubs(const char *file)
{
  xmlparser xml;

	if(!xml.Read(file)) 
  	throw std::exception("Could not load club database.");

	const xmlobject *xo;

	//Get clubs
	xo=xml.GetObject("ClubList");
	if (xo) {
		const xmlList *xl=xo->GetObjects();

		if(xl){
			xmlList::const_iterator it;
      cdb.clear();
      chash.clear();
      cdb.reserve(xl->size());
			for (it=xl->begin(); it != xl->end(); ++it) {
				if(it->Is("Club")){
					oClub c(oe);
					c.Set(&*it);
          chash[c.getId()]=cdb.size();
          cdb.push_back(c);
				}
			}
		}
  }
}

void RunnerDB::loadRunners(const char *file)
{
  string ex=string("Bad runner database. ")+file;
  int f=-1;
  _sopen_s(&f, file, _O_BINARY|_O_RDONLY, 
            _SH_DENYWR, _S_IREAD|_S_IWRITE);
 
  if (f!=-1) {
    rhash.clear();
    int len = _filelength(f);
    if (len%sizeof(RunnerDBEntry) != 0) {
      _close(f);
      throw std::exception(ex.c_str());
    }
    int nentry = len / sizeof(RunnerDBEntry);
    
    rdb.resize(nentry);
    if (rdb.empty())
      return;

    _read(f, &rdb[0], len);
    _close(f);
    for (int k=0;k<nentry;k++) {
      if (!check(rdb[k]))
        throw std::exception(ex.c_str());
    }
    
    for (int k=0;k<nentry;k++) 
      rhash[rdb[k].cardNo]=k;
  }
  else throw std::exception(ex.c_str());
}

bool RunnerDB::check(const RunnerDBEntry &rde) const
{
  if (rde.cardNo<=0 || rde.cardNo>99999999 
           || rde.name[baseNameLength-1]!=0 || rde.clubNo<0)
    return false;
  return true;
}

void RunnerDB::addCR(int card, const string &name, 
                     const string &club) 
{
  if (card) {
    pClub c=getClub(club);
    int cid;
    if (c)
      cid=c->getId();
    else {
      oClub nc(oe);
      nc.setName(club);
      cid = nc.Id = chash.empty() ? 1 :chash.rbegin()->first+1;
      addClub(nc);
    }
    addRunner(name.c_str(), cid, card);
  }
}

void RunnerDB::updateAdd(int card, const string &name, const pClub club)
{  
  pClub pc = club ? getClub(club->getId()) : 0;

  int oldid = pc ? pc->getId() : 0;
  if (club) {
    if (pc==0 || _stricmp(pc->getName().c_str(), club->getName().c_str())) {
      pc = getClub(club->getName());
      
      if(pc==0) {
        if (oldid) {
          oClub c(*club);
          oldid = c.Id = chash.empty() ? 1 :chash.rbegin()->first+1;
          addClub(c);
        }
        else {
          addClub(*club);
          oldid = club->getId();
        }
      }
    }
  }
  RunnerDBEntry *rde=getRunner(card);

  if (rde) { 
    rde->setName(name.c_str());
    rde->clubNo = oldid;
  }
  else
    addRunner(name.c_str(), oldid, card);
}

void RunnerDB::getAllNames(vector<string> &givenName, vector<string> &familyName) 
{
  givenName.reserve(rdb.size());
  familyName.reserve(rdb.size());
  for (size_t k=0;k<rdb.size(); k++) {
    string gname(rdb[k].getGivenName());
    string fname(rdb[k].getFamilyName());
    if(!gname.empty())
      givenName.push_back(gname);
    if(!fname.empty())
      familyName.push_back(fname);
  }
}
